# PACSeqAnalysisPolyA_DB.py: Alternative Poly adenylation analysis from PAC Seq data #
# Hari Krishna Y  #
# Usage: PACSeqAnalysisPolyA_DB.py <config file> #

import os, sys, glob, time, uuid,math, yaml
from pybedtools import BedTool
import pandas as pd
import numpy as np
import concurrent.futures as cf


# check strand and make bed for PolyA_DB #
def chk_strand(row,APAblock):
	if row['strand']=="+":
		row['start']=row['start']-APAblock
	if row['strand']=="-":
		row['end']=row['end']+APAblock
	return row

# Process Raw files Trip first 6 bases, adapters and other quality filters #
def process_rawfastq(d,s,mc):
		if int(mc) > 16:
			mc="16" 
		cmd="fastp -i "+s+" -a AGATCGGAAGAGC -f 6 -g -l 40 -h "+d+"/"+d.split("/")[-1]+"_trim-report.html -w "+mc+" -Q -o "+d+"/"+d.split("/")[-1]+"_prep.txt"
		print (cmd)
		os.system(cmd)
		cmd="pigz -p "+mc+" "+d+"/"+d.split("/")[-1]+"_prep.txt"
		print (cmd)
		os.system(cmd)

# bowtie2 mapping, sam to bam, bam sort and index, del intermediate files #
def mapping_bowtie2(d,mc,ref_genome):
	cmd="bowtie2 -p "+mc+" --very-sensitive-local -x "+ref_genome+' -U '+d+'/'+d.split('/')[-1]+'_prep.txt.gz -S '+d+'/'+d.split('/')[-1]+".sam"
	print (cmd)
	os.system(cmd)

def sam2bam(f):
	cmd="samtools view -bS "+f+" -o "+f.replace(".sam",".bam")
	print (cmd)
	os.system(cmd)

def sortbam(f,mc):	
	cmd="samtools sort -@ "+mc+" "+f+" "+f.replace(".bam",".sorted")
	print (cmd)
	os.system(cmd)
	os.system("rm "+f)

def indexbam(f):
	cmd="samtools index "+f
	print (cmd)
	os.system(cmd)

# bed coverage track #
def make_bg(file):
	tfp = str(uuid.uuid4())
	bamf= BedTool(file)
	bgf = bamf.genome_coverage(bg=True,strand='+')
	bgf.saveas(tfp)
	df = pd.read_csv(tfp,sep="\t",header=None,names=['Chr','Start','End','Strand'],index_col=None)
	df['Strand']="+"
	df.to_csv(tfp, sep="\t", index=False,header=None)
	# -Ve strand #
	tfn = str(uuid.uuid4())
	bgf = bamf.genome_coverage(bg=True,strand='-')
	bgf.saveas(tfn)
	df = pd.read_csv(tfn,sep="\t",header=None,names=['Chr','Start','End','Strand'],index_col=None)
	df['Strand']="-"
	df.to_csv(tfn, sep="\t", index=False,header=None)
	os.system("cat  "+tfp+" "+tfn+" > "+file.replace(".sorted.bam",".bg"))
	temp = BedTool(file.replace(".sorted.bam",".bg"))
	temp=temp.sort()
	temp.saveas(file.replace(".sorted.bam",".bg"))
	os.system("rm "+tfp+" "+tfn)

# Merge BG files 
def mergeBG(f):
	pars=f.split(",")
	# Collapse bg file #
	lines=open(pars[0],"r").readlines()
	fw=open(pars[0].replace(".bg",".bed"),"w")
	for line in lines:
		data=line.strip().split("\t")
		if "M" not in data[0] and "_" not in data[0]:
			fw.write(data[0]+"\t"+data[1]+"\t"+data[2]+"\t"+data[0]+"_"+data[1]+"_"+data[2]+"_"+data[3]+"\tN\t"+data[3]+"\n")
	fw.close()
	bed = BedTool(pars[0].replace(".bg",".bed"))
	bed=bed.sort()
	merged = bed.merge(s=True)
	merged.saveas(pars[0].replace(".bg",".bed"))

# Get count #
def UTR3_counts(f):
	file=f.split(",")[0]
	saf=f.split(",")[1]
	cmd="featureCounts -a "+saf+" -F SAF -f -O --readExtension5 0 --readExtension3 0 -M -s 0 -T 5 -o "+file.replace(".sorted.bam","")+".UTR3.Counts.txt "+file
	os.system(cmd)

# lambda function for pOvera #
def check_PoverA_OLD(row,P,A,M):
	count=0.0
	for i in range(2,len(row)):
		if row[i] >= int(A):
			count=count+1.0
		if row[i] < int(M):
			return (0)
	if count >= (float(P)*(len(row)-1)):
		return(1)

# lambda function for pOvera #
def check_PoverA(row,P,A,M):
	count=0.0
	nrow=list(row)[2:]
	for i in nrow:
		if i >= int(A):
			count=count+1.0
		if i < int(M):
			return (0)
	if count >= (float(P)*len(nrow)):
		return(1)

# pecent over minimum read count #
def PoverA(pars):
	pars[0][pars[4]]= pars[0].apply(lambda row: check_PoverA(row,pars[1],pars[2],pars[3]),axis=1)
	return (pars[0])

# Add AP type to test out #
def addAPtype(row,tdf):
	tdf=tdf[tdf['gene_id'] == row['gene_id']]
	temp=",".join(list(tdf['APtype']))
	return ",".join(list(set(temp.split(","))))

# conver Counts to Prop
def count2proplist(lst):
	tot=sum(lst)
	if tot > 0:
		for i in range(0,len(lst)):
			lst[i]=lst[i]/tot
		return lst
	else:
		temp=[]
		for i in range(0,len(lst)):
			temp.append(0)
		return temp

# Filterout non expressed genes #
def GeneFilter(gene,df,controls,cases):
	tdf=df.copy()
	tdf=tdf[tdf['gene_id']==gene]
	if tdf.shape[0] <= 1:
		return None
	ct=list(tdf[controls].mean(axis=1))
	ctn=len(ct)
	cts=sum(ct)
	tr=list(tdf[cases].mean(axis=1))
	trn=len(tr)
	trs=sum(tr)
	if cts/float(ctn) < 5 or trs/float(trn) < 5:
		return(0)
	else:
		return(1) 

# Filterout low proportion APA sites #
def PropFilter(gene,df,controls,cases):
	tdf=df.copy()
	tdf=tdf[tdf['gene_id']==gene]
	#print(tdf.shape[0])
	if tdf.shape[0] <= 1:
		return None
	ct=count2proplist(list(tdf[controls].mean(axis=1)))
	tr=count2proplist(list(tdf[cases].mean(axis=1)))
	passlist=[]
	for i in range (0,len(ct)):
		if ct[i] <=0.1 and tr[i] <=0.1:
			pass
		else:
			passlist.append(i)
	return tdf.iloc[passlist,:]

# Add control propDiff  #
def globalCR(row,tdf,controls,cases):
	tdf=tdf[tdf['gene_id']==row['gene_id']]
	if tdf.shape[0] <= 1:
		return "NA"
	ct=count2proplist(list(tdf[controls].mean(axis=1)))
	if "+" in list(set((list(tdf['Strand'])))):
		try:
			return(ct[-1]-ct[0])
		except:
			return(ct[-1]/0.01)
	if "-" in list(set((list(tdf['Strand'])))):
		try:
			return(ct[0]-ct[-1])
		except:
			return(ct[0]/0.01)

# Add treated propDiff  #
def globalTR(row,tdf,controls,cases):
	tdf=tdf[tdf['gene_id']==row['gene_id']]
	if tdf.shape[0] <= 1:
		return "NA"
	tr=count2proplist(list(tdf[cases].mean(axis=1)))
	if "+" in list(set((list(tdf['Strand'])))):
		try:
			return(tr[-1]-tr[0])
		except:
			return(tr[-1]/0.01)
	if "-" in list(set((list(tdf['Strand'])))):
		try:
			return(tr[0]-tr[-1])
		except:
			return(tr[0]/0.01)

# Add global propDiff  #
def globalPD(row,tdf,controls,cases):
	tdf=tdf[tdf['gene_id']==row['gene_id']]
	if tdf.shape[0] <= 1:
		return "NA"
	ct=count2proplist(list(tdf[controls].mean(axis=1)))
	tr=count2proplist(list(tdf[cases].mean(axis=1)))
	rank=[]
	for i in range(0,len(ct)-1):
		rank.append(1-((1/(len(ct)-1)*i)))
	if "+" in list(set((list(tdf['Strand'])))):
		rank.reverse()
		ct.reverse()
		tr.reverse()
	C=0.0
	T=0.0
	for i in range(0,len(ct)-1):
		C=C+ct[i]*rank[i]
		T=T+tr[i]*rank[i]
	return(T-C)

################################# MAIN #################################

# Parse YAML file #
with open(sys.argv[1], 'r') as f:
	args = yaml.load(f)

# Read arguments #
baseDir=os.path.abspath(args["APSeq"]["baseDir"])
outDir=os.path.abspath(args["APSeq"]["outDir"])
mc=args["APSeq"]["mc"]
ref_genome= args["APSeq"]["ref_genome"]
ref_fasta= os.path.abspath(args["APSeq"]["ref_fasta"])
ref_bed=os.path.abspath(args["APSeq"]["ref_bed"])
ref_polyA=os.path.abspath(args["APSeq"]["ref_polyA"])
anchor=int(args["APSeq"]["anchor"])
APAblock=int(args["APSeq"]["APAblock"])
mddb=int(args["APSeq"]["mddb"])
md=int(args["APSeq"]["md"])
controls=args["APSeq"]["ct"].split(",")
cases=args["APSeq"]["tr"].split(",")
pA=args["APSeq"]["pA"]
rc=int(args["APSeq"]["rc"])
M=args["APSeq"]["M"]
fkey=args["APSeq"]["fkey"]

# Log  #
localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile = open(outDir+"/"+fkey+'.PACSeqAnalysisPolyA_DB.log.txt','w')
logfile.write('# Satrting APSeq pipeline: '+localdate+' at: ' + localtime+' \n')

# check for OUT directory #
if (os.path.isdir(outDir)):
	pass
else:
	os.system("mkdir "+outDir)
outDir=outDir+"/"

# Extract PolyA_DB APA sites #

df=pd.read_csv(ref_polyA,sep="\t",index_col=None,header=None,names=['chr','start','end','gene','PAtype','strand'])
df=df[df['gene']!='na']
df=df.apply(lambda row: chk_strand(row,APAblock),axis=1)
df.to_csv(outDir+fkey+".APSites.bed",sep="\t",index=None,header=None)
bed = BedTool(outDir+fkey+".APSites.bed")
bed=bed.sort()
merged = bed.merge(s=True,c=[4,5],o="distinct",d=mddb)
merged.saveas(outDir+fkey+".APSites.bed")
df=pd.read_csv(outDir+fkey+".APSites.bed",sep="\t",index_col=None,header=None,names=['chr','start','end','strand','gene','PAtype'])
df=df[['chr','start','end','gene','PAtype','strand']]
df=df[~df['chr'].str.contains('_')]
df=df[~df['gene'].str.contains(',')]
df.to_csv(outDir+fkey+".APSites.bed",sep="\t",index=None,header=None)
# Log # 
localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished extracting AP sites from PolyA_DB: '+localdate+' at: ' + localtime+' \n')

samples = glob.glob (baseDir+'/*.fastq.gz')


# Processing Data -Reads #
for s in samples:
	d=outDir+s.split("/")[-1].replace(".fastq.gz","")
	os.system("mkdir "+d)
	process_rawfastq(d,s,mc)
	localdate = time.strftime('%a %m/%d/%Y')
	localtime = time.strftime('%H:%M:%S')
	logfile.write('# Finished fastp: '+localdate+' at: ' + localtime + 'for '+d+' \n')
	os.system("rm *fastp.json")
	# Mapping to Reference # 
	mapping_bowtie2(d,mc,ref_genome)
	localdate = time.strftime('%a %m/%d/%Y')
	localtime = time.strftime('%H:%M:%S')
	logfile.write('# Finished mapping on '+localdate+' at: ' + localtime + 'for '+d+' \n')

# sam2bam #
files=glob.glob(outDir+"*/"+"*.sam")
with cf.ProcessPoolExecutor(max_workers=int(mc)) as executor:
	result = list(executor.map(sam2bam, files))
os.system("rm "+outDir+"*/"+"*.sam")
localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished converting sam 2 bam on '+localdate+' at: ' + localtime + ' ##\n')

# sort bam #
files=glob.glob(outDir+"*/"+"*.bam")
for f in files:
	sortbam(f,mc)
localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished sorting bams on '+localdate+' at: ' + localtime + ' ##\n')

# indexbam #
files=glob.glob(outDir+"*/"+"*.sorted.bam")
with cf.ProcessPoolExecutor(max_workers=int(mc)) as executor:
	result = list(executor.map(indexbam, files))
localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished indexing bams on '+localdate+' at: ' + localtime + ' ##\n')

# Del processed reads #
os.system("rm "+outDir+"*/"+"*_prep.txt.gz")

# MultiQC #
os.system("multiqc "+outDir+" -o "+outDir+fkey+"_MultiQCReport")

# Make feature file #
files=glob.glob(outDir+"*/"+"*.sorted.bam")
with cf.ProcessPoolExecutor(max_workers=int(mc)) as executor:
	result = list(executor.map(make_bg, files))
localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finifhed generating ged graph tracks: '+localdate+' at: ' + localtime +' \n')

# Merge sample wise BG files #

files=glob.glob(outDir+"*/"+"*.bg")
passlist=[]
for file in files:
	passlist.append(file)
with cf.ProcessPoolExecutor(max_workers=int(mc)) as executor:
	result = list(executor.map(mergeBG, passlist))
localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished merging individial feature bed grapg files: '+localdate+' at: ' + localtime + ' \n')

# Make union features #
files=glob.glob(outDir+"*/"+"*.bed")
file=" ".join(files)
os.system("cat "+file+" > "+outDir+fkey+"_Jumbo.bed")
df = pd.read_csv(outDir+fkey+"_Jumbo.bed",sep="\t",header=None,names=['Chr','Start','End','Strand'],index_col=None)
df['N1']="N1"
df['N2']="N2"
df=df[['Chr','Start','End','N1','N2','Strand']]
df.to_csv(outDir+fkey+"_Jumbo.bed", sep="\t", index=False,header=None)
pooled = BedTool(outDir+fkey+'_Jumbo.bed')
pooled=pooled.sort()
pooled.saveas(outDir+fkey+"_Jumbo.bed")

# Merging a file #
merged_bg = pooled.merge(s=True,d=md)
merged_bg.saveas(outDir+fkey+"_merged.features.bed")

# Process DF #
df = pd.read_csv(outDir+fkey+"_merged.features.bed",sep="\t",header=None,names=['Chr','Start','End','Strand'],index_col=None)
nrow=df.shape[0]
names=[]
for n in range(1,(nrow+1)):
	names.append("3UTR_"+str(n))
df['GeneID']=names

# make features bed #
df=df[['Chr','Start','End','GeneID','GeneID','Strand']]
df.to_csv(outDir+fkey+"_All3UTRs.bed", sep="\t", index=False,header=None)
os.system("rm "+outDir+fkey+"_Jumbo.bed "+outDir+fkey+"_merged.features.bed")

localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finifhed generating features bed file on '+localdate+' at: ' + localtime +' ##\n')

# Intersect annotated APA sites with features #
os.system("bedtools intersect -a "+outDir+fkey+".APSites.bed -b "+outDir+fkey+"_All3UTRs.bed -s -wo > "+outDir+fkey+"_AP.Distance.txt")

# Write distance #
lines=open(outDir+fkey+"_AP.Distance.txt","r").readlines()
fw=open(outDir+fkey+"_AP.Distance.txt","w")
for line in lines:
	data=line.strip().split("\t")
	if data[5] == "+" and int(data[12]) >= anchor:
		fw.write("\t".join(data)+"\t"+str(int(data[8])-int(data[7]))+"\n")
	if data[5] == "-" and int(data[12]) >= anchor:
		fw.write("\t".join(data)+"\t"+str(int(data[8])-int(data[7]))+"\n")
fw.close()

localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished computing distance and ovrlap with annotated AP sites: '+localdate+' at: ' + localtime + ' \n')

# Make SAF file #
df=pd.read_csv(outDir+fkey+"_AP.Distance.txt",sep="\t",header=None,index_col=None,names=['Chr1','Start1','End1','gene_id','isoform','Strand1','Chr','Start','End','UTRID1','isoform1','Strand','OV','DS'])
df=df[['Chr','Start','End','gene_id','isoform','Strand']]
df=df.drop_duplicates(subset=['Chr','Start','End','Strand'])
df.to_csv(outDir+fkey+"_APtype.txt", sep="\t", index=False,header=None)
df['GeneID']=df['gene_id']+"@"+df['isoform']
df=df[['GeneID','Chr','Start','End','Strand']]
df.to_csv(outDir+fkey+"_UTR3.saf", sep="\t", index=False,header=None)

# Get UTR counts #
files=glob.glob(outDir+"*/"+"*.sorted.bam")

for i in range(0,len(files)):
	files[i]=files[i]+","+outDir+fkey+"_UTR3.saf"
with cf.ProcessPoolExecutor(max_workers=int(mc)) as executor:
	result = list(executor.map(UTR3_counts, files))
os.system("rm "+outDir+fkey+"_UTR3.saf "+outDir+"*/*.txt.summary")

localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished feature counts: '+localdate+' at: ' + localtime + ' \n')

# Make data matrix #
samples = glob.glob (outDir+"*/"+'*.UTR3.Counts.txt')
df=pd.read_csv(samples[0],comment='#',sep="\t",index_col=None)
df.columns=['Geneid','Chr','Start','End','Strand','Length',samples[0].split("/")[-1].replace(".UTR3.Counts.txt","")]
for s in samples[1:]:
	dftemp=pd.read_csv(s,comment='#',sep="\t",index_col=None)
	df[s.split("/")[-1].replace(".UTR3.Counts.txt","")]=dftemp[s.replace('.UTR3.Counts.txt','.sorted.bam')].copy()

df["feature_id"]=df["Chr"]+"_"+df["Start"].astype(str)+"_"+df["End"].astype(str)+"_"+df["Strand"]
df[['gene_id','isoform']]= df.Geneid.str.split('@', expand=True)
design=["feature_id",'gene_id']+controls+cases
df=df[design]
df.to_csv(outDir+fkey+"_3UTRCountMatrix.txt",sep="\t",index=False)

localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished making data matrix : '+localdate+' at: ' + localtime + ' \n')

# PoverA Filtering #
cdf=df[["feature_id",'gene_id']+controls]
tdf=df[["feature_id",'gene_id']+cases]
passlist=[[cdf,str(pA),str(rc),str(M),"C"],[tdf,str(pA),str(rc),str(M),"T"]]
with cf.ProcessPoolExecutor(max_workers=int(mc)) as executor:
	result = list(executor.map(PoverA, passlist))
df=pd.merge(result[0],result[1],on=['feature_id','gene_id'],how='outer')
df=df[design+['C']+['T']]
df=df.fillna(0)
df=df[(df['C'] > 0) | (df['T'] > 0)]
df=df.drop(columns=['C','T'])
df.to_csv(outDir+fkey+"_3UTRCountMatrix.Fil.txt",sep="\t",index=False)

# get gene level counts #
gdf=df.drop(columns=['feature_id']).copy()
gdf[gdf.columns[1:]]=gdf[gdf.columns[1:]].apply(pd.to_numeric)
gdf = gdf.groupby(['gene_id'], as_index=False, sort=False).aggregate('sum')
#gdf=gdf[(df.gene_id != ".")]
gdf.to_csv(outDir+fkey+".GeneCounts.txt",sep="\t", index=False)

#Proportion Filtering #
genes=list(set(list(df['gene_id'])))
fdf=pd.DataFrame()
for gene in genes:
	t=PropFilter(gene,df,controls,cases)
	try:
		if (t.shape[0] >0):
			fdf=fdf.append(t, ignore_index=True)
	except:
		pass
fdf.to_csv(outDir+fkey+"_3UTRCountMatrix.PFil.txt",sep="\t",index=False)

# Gene Expression Filter #
df=pd.read_csv(outDir+fkey+"_3UTRCountMatrix.PFil.txt",comment='#',sep="\t",index_col=None)
fdf=pd.DataFrame()
for gene in genes:
	if GeneFilter(gene,df,controls,cases):
		fdf=fdf.append(df[df['gene_id']==gene], ignore_index=True)

fdf.to_csv(outDir+fkey+"_3UTRCountMatrix.PGFil.txt",sep="\t",index=False)

localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished filtering count matrix: '+localdate+' at: ' + localtime + ' \n')


# write design file #
fw=open(outDir+fkey+"_APDesign.txt","w")
fw.write("sample_id\tgroup\n")
for s in controls:
	fw.write(s+"\tCON\n")
for s in cases:
	fw.write(s+"\tTRT\n")
fw.close()


# DM-test #
os.system("Rscript PATest.r "+outDir+fkey+"_3UTRCountMatrix.PGFil.txt "+outDir+fkey+"_APDesign.txt "+outDir+fkey+"_AP.Changes.txt")

localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('# Finished APSeq DMtest : '+localdate+' at: ' + localtime+' \n')

# add APtpe #
df=pd.read_csv(outDir+fkey+"_AP.Changes.txt",sep="\t",header=0,index_col=None)
df1=pd.read_csv(outDir+fkey+"_APtype.txt",sep="\t",header=None,index_col=None,names=['Chr','Start','End','gene_id','APtype','Strand'])
df2=pd.read_csv(outDir+fkey+"_3UTRCountMatrix.PGFil.txt",sep="\t",header=0,index_col=None)
df2[['Chr','Start','End','Strand']]=df2['feature_id'].str.split('_',expand=True)
df2['Start']=df2['Start'].astype(int)
df2['End']=df2['End'].astype(int)
df2=df2.merge(df1,on=['Chr','Start','End','Strand','gene_id'],how='inner')
df2=df2[['Chr','Start','End','gene_id','APtype','Strand']+controls+cases]
df2.to_csv(outDir+fkey+"_3UTRCountMatrix.PGFil.txt",sep="\t",index=None)
df['APtype']=df.apply(lambda row: addAPtype(row,df2),axis=1)

# add control magnitude #
#df['CR']=df.apply(lambda row: globalCR(row,df2,controls,cases),axis=1)
# add treat magnitude #
#df['TR']=df.apply(lambda row: globalTR(row,df2,controls,cases),axis=1)
# add switch magnitude #
df['SM']=df.apply(lambda row: globalPD(row,df2,controls,cases),axis=1)
# Clean #
os.system("rm "+outDir+fkey+"*_3UTR* "+outDir+fkey+"*_AP* "+outDir+fkey+"*.bed "+outDir+fkey+"*.GeneCounts.txt")
# Output #
df.to_csv(outDir+fkey+"_AP.Changes.txt",sep="\t",index=None)

localdate = time.strftime('%a %m/%d/%Y')
localtime = time.strftime('%H:%M:%S')
logfile.write('############## Finished APSeq pipeline: '+localdate+' at: ' + localtime+' ##############\n')
logfile.close()
